/* Copyright 2021 Jesse Haber-Kucharsky */
/* SPDX-License-Identifier: GPL-3.0-only */

#include "list.hpp"

#include <pretty.hpp>

#include <algorithm>
#include <iostream>
#include <iterator>
#include <memory>
#include <string>
#include <type_traits>
#include <utility>
#include <variant>
#include <vector>

class Sexp final {
 public:
  Sexp() : _data{List<std::shared_ptr<Sexp const>>{}} {
  }
  explicit Sexp(std::string s) : _data{std::move(s)} {
  }
  explicit Sexp(List<std::shared_ptr<Sexp const>> xs) : _data{std::move(xs)} {
  }

  pretty::Doc prettify() const {
    using namespace pretty;

    if (auto *atom = std::get_if<std::string>(&_data); atom) {
      return text(*atom);
    }

    auto const &xs = std::get<List<std::shared_ptr<Sexp const>>>(_data);
    std::vector<Doc> ds{};

    std::transform(
        xs.cbegin(), xs.cend(), std::back_inserter(ds), [](std::shared_ptr<Sexp const> const &xp) {
          return xp->prettify();
        });

    return group(char_('(') + nest(1, fill(ds.begin(), ds.end())) + char_(')'));
  }

 private:
  std::variant<std::string, List<std::shared_ptr<Sexp const>>> _data;
};

template <typename... Xs>
Sexp make(Xs const &...xs) {
  return Sexp{List<std::shared_ptr<Sexp const>>{std::make_shared<Sexp const>(xs)...}};
}

int main() {
  auto xs = make("a", "b", "c", "d");
  xs = make("xxxxxxx", make(xs, xs, xs, xs));
  auto d = xs.prettify();
  std::cout << d.fit_to(15) << std::endl;
}
