{
  description = "A pretty-printing library for C++";

  inputs = {
    nixpkgs.url = "github:NixOS/nixpkgs/nixos-21.05";
  };

  outputs = { self, nixpkgs }:
    let
      version = "1.0.1";
      system = "x86_64-linux";
      pkgs = import nixpkgs { inherit system; };
      stdenv = pkgs.gcc11Stdenv;
      emacsWithPackages = (pkgs.emacsPackagesFor pkgs.emacs-nox).emacsWithPackages;

      emacs = emacsWithPackages (epkgs: [
        epkgs.melpaStablePackages.htmlize
      ]);
    in
    {
      devShell.${system} = pkgs.mkShell.override { inherit stdenv; } {
        buildInputs = [
          emacs
          # For `clang-format` and `clangd`.
          pkgs.clang_12
          pkgs.cmake
          pkgs.gtest
          pkgs.ninja
        ];
      };

      packages.${system} = {
        pretty = stdenv.mkDerivation {
          name = "pretty";
          src = self;
          nativeBuildInputs = [ pkgs.cmake pkgs.gtest ];
          cmakeFlags = [ "-DPretty_TESTS=yes" ];
          doCheck = true;
        };

        doc = stdenv.mkDerivation {
          name = "pretty-doc";
          src = self;

          postPatch = ''
            substituteInPlace overview.org \
                --replace '%%VERSION%%' "${version}";
          '';

          buildInputs = [ emacs ];
          buildPhase = "make doc";

          installPhase = ''
            mkdir -p $out
            cp overview.html $out/index.html
            cp api.html development.html implementation.html $out/
          '';
        };
      };

      defaultPackage.${system} = self.packages.${system}.pretty;

      checks.${system} = {
        compare = import ./test_compare {
          inherit (pkgs) ghc;
          pretty = self.defaultPackage.${system};
          inherit stdenv;
        };

        dist = import ./test_dist {
          inherit (pkgs) cmake;
          pretty = self.defaultPackage.${system};
          inherit stdenv;
        };
      };
    };
}
