module Main where

import Prelude hiding ((<>))

infixr 5 :<|>

infixr 6 :<>

infixr 6 <>

data DOC
  = NIL
  | DOC :<> DOC
  | NEST Int DOC
  | TEXT String
  | LINE
  | DOC :<|> DOC

data Doc
  = Nil
  | String `Text` Doc
  | Int `Line` Doc

nil :: DOC
nil = NIL

(<>) :: DOC -> DOC -> DOC
x <> y = x :<> y

nest :: Int -> DOC -> DOC
nest i x = NEST i x

text :: String -> DOC
text s = TEXT s

line :: DOC
line = LINE

group :: DOC -> DOC
group x = flatten x :<|> x

flatten :: DOC -> DOC
flatten NIL = NIL
flatten (x :<> y) = flatten x :<> flatten y
flatten (NEST i x) = NEST i (flatten x)
flatten (TEXT s) = TEXT s
flatten LINE = TEXT " "
flatten (x :<|> _) = flatten x

layout :: Doc -> String
layout Nil = ""
layout (s `Text` x) = s ++ layout x
layout (i `Line` x) = '\n' : copy i ' ' ++ layout x

copy :: Int -> Char -> String
copy i x = [x | _ <- [1 .. i]]

best :: Int -> Int -> DOC -> Doc
best w k x = be w k [(0, x)]

be :: Int -> Int -> [(Int, DOC)] -> Doc
be _ _ [] = Nil
be w k ((_, NIL) : z) = be w k z
be w k ((i, x :<> y) : z) = be w k ((i, x) : (i, y) : z)
be w k ((i, NEST j x) : z) = be w k ((i + j, x) : z)
be w k ((_, TEXT s) : z) = s `Text` (be w (k + length s) z)
be w _ ((i, LINE) : z) = i `Line` (be w i z)
be w k ((i, x :<|> y) : z) = better w k (be w k ((i, x) : z)) (be w k ((i, y) : z))

better :: Int -> Int -> Doc -> Doc -> Doc
better w k x y = if fits (w - k) x then x else y

fits :: Int -> Doc -> Bool
fits w _ | w < 0 = False
fits _ Nil = True
fits w (s `Text` x) = fits (w - length s) x
fits _ (_ `Line` _) = True

pretty :: Int -> DOC -> String
pretty w x = layout (best w 0 x)

-- Utility functions.

(<+>) :: DOC -> DOC -> DOC
x <+> y = x <> text " " <> y

(</>) :: DOC -> DOC -> DOC
x </> y = x <> line <> y

folddoc :: (DOC -> DOC -> DOC) -> [DOC] -> DOC
folddoc _ [] = nil
folddoc _ [x] = x
folddoc f (x : xs) = f x (folddoc f xs)

(<+/>) :: DOC -> DOC -> DOC
x <+/> y = x <> (text " " :<|> line) <> y

fillwords :: String -> DOC
fillwords = folddoc (<+/>) . map text . words

main :: IO ()
main = do
  s <- getContents
  putStrLn (pretty 72 (fillwords s))
